#include "application.h"

#include <iostream>
#include <cassert>
#include <cstdio>

using namespace std;

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "matrix.h"
void set_pixel(int x, int y, float col[3]);

application::application() 
    : niter(20), md(RENDER), img(NULL)
{
}

application::~application()
{
    if (img) delete[] img;
}

// triggered once after the OpenGL context is initialized
void application::init_event()
{
    cout << "LMB: drag to select a region of the fractal to zoom in on" << endl;
    cout << "RMB: zoom out to previous region" << endl;
    glClearColor(0.0, 0.0, 0.0, 0); 

    // initial complex plane mapping to viewport
    cb.set(-2.0, -1.0, 1, 1.0);
}

// maps the viewport to the the complex plane
complex_number application::viewport_to_complex_plane(
    int i, int j
)
{
    float vp_aspect = width/(float)height;
    float cb_width = abs(cb.x1 - cb.x0);
    float cb_height = cb_width/vp_aspect;
    float rem = abs(cb.y1 - cb.y0)-cb_height;
    float vp_width = width;
    float vp_height = height;
    
    float a, b;
    a = cb_width/vp_width;
    b = cb_height/vp_height;

    return complex_number(a*i+cb.x0, b*j+cb.y0+rem/2);
}

// triggered each time the application needs to redraw
void application::draw_event()
{

    if (md == SELECTION) {
        //
        // This code is used to draw the 'zoom' box over the image. Because
        // rendering the fractal is expensive, and the box needs to be redrawn
        // as the user drags the mouse, this code draws a stored rendering 
        // of the fractal instead.
        //
        glDrawPixels(width, height, GL_RGB, GL_BYTE, img);


        //
        // draw the box for zooming in
        //
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
        glColor4f(.5,.5,0,0.3);
        glBegin(GL_POLYGON);
            glVertex2f(sb.x0, sb.y0);
            glVertex2f(sb.x0, sb.y1);
            glVertex2f(sb.x1, sb.y1);
            glVertex2f(sb.x1, sb.y0);
        glEnd();
        glColor3f(1,1,0);
        glBegin(GL_LINE_LOOP);
            glVertex2f(sb.x0, sb.y0);
            glVertex2f(sb.x0, sb.y1);
            glVertex2f(sb.x1, sb.y1);
            glVertex2f(sb.x1, sb.y0);
        glEnd();
    } else if (md == RENDER) {
        for (int i = 0; i < width; ++i) {
            for (int j = 0; j < height; ++j) {
                complex_number c = viewport_to_complex_plane(i,j);

                //
                //
                //
                // ADD YOUR CODE HERE
                //
                //
                //
            }
        }
    }
}

// triggered when mouse is clicked
void application::mouse_click_event(
    mouse_button button, mouse_button_state button_state, 
    int x, int y
    )
{
    // store point if user just depressed the left button
    if (button == LEFT_BUTTON && button_state == BUTTON_DOWN)
    {
        complex_number z = viewport_to_complex_plane(x, y);

        sb.x0 = x; sb.y0 = y;
        sb.x1 = x; sb.y1 = y;

        z.get(cb_new.x0, cb_new.y0);

        // store the current rendered fractal as an image array
        // (to keep having to regenerate the fractal when drawing the zoom box)
        if (img) { 
            delete img; 
            img = NULL;
        }
        img = new char[width*height*3*8];
        glReadPixels(0,0,width,height,GL_RGB, GL_BYTE, img);
        
        md = SELECTION;
        glutPostRedisplay();
    }
    if (button == LEFT_BUTTON && button_state == BUTTON_UP) {
        // only zoom in the zoom box has a positive area
        if (sb.x0 == x || sb.y0 == y) {
            return;
        }

        complex_number z = viewport_to_complex_plane(x, y);
        z.get(cb_new.x1, cb_new.y1);

        if (cb_new.x0 > cb_new.x1) {
            swap(cb_new.x0, cb_new.x1);
        }
        if (cb_new.y0 > cb_new.y1) {
            swap(cb_new.y0, cb_new.y1);
        }

        // push current complex boundary into tack
        zoom_history.push_back(cb);

        // update the complex boundary
        cb = cb_new;

        md = RENDER;
        glutPostRedisplay();
        
        cout << "zooming in: (" 
            << cb.x0 << ", " << cb.y0 << ", " << cb.x1 << ", " << cb.y1 << ")" 
            << endl;
    }
    
    if (button == RIGHT_BUTTON && button_state == BUTTON_DOWN) {
        // prevents the user zooming out further than the
        // original zoom level
        if (!zoom_history.empty())
        {
            // goes back to the previous zoom level
            cb_new = zoom_history[zoom_history.size()-1];
            zoom_history.pop_back();
            cb.set(cb_new.x0, cb_new.y0, cb_new.x1, cb_new.y1);
            glutPostRedisplay();
        }
        cout << "zooming out: (" 
            << cb.x0 << ", " << cb.y0 << ", " << cb.x1 << ", " << cb.y1 << ")" 
            << endl;
    }
}
    
// triggered when mouse button is held down and the mouse is
// moved
void application::mouse_move_event(
    int x, int y
    )
{
    if (md == SELECTION) {
        int sign = y > sb.y0 ? 1 : -1;
        
        // set the second endpoint of the viewport aspect-ratio perserving zoom box
        sb.x1 = x; sb.y1 = (float)fabs(sb.x1-sb.x0)*height/width*sign+sb.y0;
        glutPostRedisplay();
    }
}

// triggered when a key is pressed on the keyboard
void application::keyboard_event(unsigned char key, int x, int y)
{
    // rerender
    if (key == '=') {
        niter += 5;
        glutPostRedisplay();
    } else if (key == '-') {
        niter = max(0, niter-5);
        glutPostRedisplay();
    } else if (key == 'q') {
        exit(0);
    }
}

void set_pixel(int x, int y, float col[3])
{
    glBegin(GL_POINTS);
    glColor3fv(col);
    glVertex2f(x, y);
    glEnd();
}

